@@ -17,13 +17,13 @@
  *   "FAILED" may not be used to test an OLE return code.  Use SUCCEEDED
  *   instead.
  *
- * $Header: /usr/build/vile/vile/RCS/w32ole.cpp,v 1.12 1999/06/07 21:56:24 cmorgan Exp $
+ * $Header: /usr/build/vile/vile/RCS/w32ole.cpp,v 1.30 2008/06/01 20:07:04 tom Exp $
  */
 
-#include <windows.h>
+#include "w32vile.h"
+
 #include <ole2.h>
 #include <comutil.h>
-#include <assert.h>
 #include <ctype.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -38,10 +38,29 @@ extern "C"
 #include "w32reg.h"
 
 #include <initguid.h>
+
+#if (_MSC_VER >= 1200) && (_MSC_VER < 1300)
+#define VC6_ADDIN 1             /* Visual C++ 6.0 */
+#else
+#define VC6_ADDIN 0
+#endif
+
+#if (_MSC_VER >= 1300) && (_MSC_VER < 1400)
+#define VC7_ADDIN 1             /* Visual C++ 7.0 */
+#else
+#define VC7_ADDIN 0
+#endif
+
+#if VC6_ADDIN
 #include <objmodel/appguid.h>
 #include <objmodel/textguid.h>
 #include <objmodel/appauto.h>
 #include <objmodel/textauto.h>
+#endif
+
+#if VC7_ADDIN
+#include <objbase.h>
+#endif
 
 static size_t   ansibuf_len,   /* scaled in bytes   */
                 olebuf_len;    /* scaled in wchar_t */
@@ -58,16 +77,37 @@ static DWORD    redirect_tid, winvile_tid;   // tid == thread id
 extern "C"
 {
 
+#if OPT_TRACE
+#undef TRACE
+#define Trace MyTrace
+static void
+Trace(const char *fmt, ...)
+{
+    FILE *fp = fopen("c:\\temp\\w32ole.log", "a");
+    if (fp != 0)
+    {
+        va_list ap;
+        va_start(ap, fmt);
+        vfprintf(fp, fmt, ap);
+        va_end(ap);
+        fclose(fp);
+    }
+}
+#define TRACE(params) Trace params
+#else
+#define OPT_TRACE 0
+#define TRACE(params) /* nothing */
+#endif
+
 int
 oleauto_init(OLEAUTO_OPTIONS *opts)
 {
-    char           *dummy = NULL;
     HRESULT        hr;
     LPCLASSFACTORY pcf = NULL;
 
     olebuf_len = ansibuf_len = 512;
-    ansibuf    = (char *) malloc(ansibuf_len);
-    olebuf     = (OLECHAR *) malloc(olebuf_len * sizeof(OLECHAR));
+    ansibuf    = typeallocn(char, ansibuf_len);
+    olebuf     = typeallocn(OLECHAR, olebuf_len);
     if (! (olebuf && ansibuf))
         return (FALSE);
 
@@ -160,13 +200,76 @@ oleauto_exit(int code)
             RevokeActiveObject(dwRegisterActiveObj, NULL);
         OleUninitialize();
     }
+#ifdef VILE_ERROR_ABORT
+#undef ExitProgram              /* use VILE_ERROR_ABORT code in main.c */
+    ExitProgram(code);
+#else
     exit(code);
+#endif
 }
 
 } /* Extern "C" */
 
 /* ----------------------- C++ Helper Functions --------------------- */
 
+#if OPT_TRACE
+static char *visible_wcs(OLECHAR *source)
+{
+    static char *result = 0;
+    static unsigned len = 0;
+    char *target;
+    unsigned need = 0;
+    unsigned n;
+    if (source != 0)
+    {
+	for (n = 0; source[n] != 0; ++n)
+	    ;
+	need = 10 * (n + 2);
+	if (need > len)
+	{
+	    len = need;
+            result = (char *) realloc(result, need * sizeof(OLECHAR));
+	}
+	for (n = 0, target = result; source[n] != 0; ++n)
+	{
+	    switch (source[n])
+	    {
+	    case ' ':
+		strcpy(target, "\\s");
+		break;
+	    case '\\':
+		strcpy(target, "\\\\");
+		break;
+	    case '\n':
+		strcpy(target, "\\n");
+		break;
+	    case '\r':
+		strcpy(target, "\\r");
+		break;
+	    case '\t':
+		strcpy(target, "\\t");
+		break;
+	    case '\f':
+		strcpy(target, "\\f");
+		break;
+	    default:
+		if (source[n] >= ' '&& source[n] < 256)
+		{
+		    sprintf(target, "%c", (char) source[n]);
+		}
+		else
+		{
+		    sprintf(target, "\\%03o", (unsigned char) source[n]);
+		}
+		break;
+	    }
+	    target += strlen(target);
+	}
+    }
+    return result;
+}
+#endif
+
 /*
  * Quick & Dirty ANSI/Unicode conversion routines.  These routines use a
  * dynamic buffer to hold the converted string so it may be any arbitrary
@@ -175,22 +278,31 @@ oleauto_exit(int code)
  * used/copied before the conversion routine is called again.
  *
  */
-#if (! defined(_UNICODE) || defined(UNICODE))
+#if !(defined(_UNICODE) || defined(UNICODE))
+
 static char *
 ConvertToAnsi(OLECHAR *szW)
 {
     size_t len;
 
     len = wcstombs(NULL, szW, 1) + 1;
-    if (len > ansibuf_len)
+    if ((len + 1) > ansibuf_len)
     {
         if (ansibuf)
             free(ansibuf);
         ansibuf_len = ansibuf_len * 2 + len;
-        if ((ansibuf = (char *) malloc(ansibuf_len)) == NULL)
+        if ((ansibuf = typeallocn(char, ansibuf_len)) == NULL)
             return (ansibuf);  /* We're gonna' die */
     }
-    wcstombs(ansibuf, szW, len);
+    if (len > 0)
+    {
+	wcstombs(ansibuf, szW, len);
+    }
+    else
+    {
+	*ansibuf = '\0';
+    }
+    TRACE(("ConvertToAnsi %d->%d:%s\n", len, strlen(ansibuf), ansibuf));
     return (ansibuf);
 }
 
@@ -200,15 +312,17 @@ ConvertToUnicode(char *szA)
     size_t len;
 
     len = strlen(szA) + 1;
+    TRACE(("ConvertToUnicode %d:%s\n", strlen(szA), szA));
     if (len > olebuf_len)
     {
         if (olebuf)
             free(olebuf);
         olebuf_len = olebuf_len * 2 + len;
-        if ((olebuf = (OLECHAR *) malloc(olebuf_len * sizeof(OLECHAR))) == NULL)
+        if ((olebuf = typeallocn(OLECHAR, olebuf_len)) == NULL)
             return (olebuf);  /* We're gonna' die */
     }
     mbstowcs(olebuf, szA, len);
+    TRACE(("Result %s\n", visible_wcs(olebuf)));
     return (olebuf);
 }
 #endif
@@ -239,11 +353,6 @@ LoadTypeInfo(ITypeInfo **pptinfo, REFCLSID clsid)
 static int
 ole_err_to_msgline(HRESULT hr, char *msg)
 {
-#define  PASS_HIGH(c)        ((c) <= print_high && (c) >= print_low)
-#define  _SPC_               ' '
-#define  _TAB_               '\t'
-#define  _TILDE_             '~'
-
     char          *lclbuf = NULL, tmp[256], outstr[256];
     unsigned char *src, *dst;
     int           print_high, print_low;
@@ -275,11 +384,6 @@ ole_err_to_msgline(HRESULT hr, char *msg)
     mlforce(outstr);
     LocalFree(lclbuf);
     return (FALSE);
-
-#undef PASS_HIGH
-#undef _SPC_
-#undef _TAB_
-#undef _TILDE_
 }
 
 /* ------------------------ Class Factory -------------------------- */
@@ -379,17 +483,16 @@ vile_oa::Create(vile_oa **ppvile, BOOL visible)
 
     // Name
     tmp = TO_OLE_STRING(prognam);
-    if (! (tmp && (pvile->m_bstrName = SysAllocString(tmp))))
+    if (! (tmp && (pvile->m_bstrName = SysAllocString(tmp)) != 0))
         return E_OUTOFMEMORY;
 
     // Load type information from type library.
     hr = LoadTypeInfo(&pvile->m_ptinfo, IID_IVileAuto);
     if (! SUCCEEDED(hr))
     {
-        MessageBox(pvile->m_hwnd,
-"Cannot load type library.\r\r Register type info using winvile's '-Or' command line switch",
-                   prognam,
-                   MB_OK | MB_ICONSTOP);
+        w32_message_box(pvile->m_hwnd,
+			"Cannot load type library.\r\r Register type info using winvile's '-Or' command line switch",
+			MB_OK | MB_ICONSTOP);
         return (hr);
     }
 
@@ -512,18 +615,15 @@ vile_oa::get_FullName(BSTR *pbstr)
         char    *cp, key[NFILEN];
         HKEY    hk;
         long    rc;
-        DWORD   regtype = REG_SZ;
         OLECHAR *tmp;
-        BYTE    value[NFILEN * 2];
-        ULONG   val_len;
+        char    value[NFILEN * 2];
 
         *pbstr  = NULL;
-        val_len = sizeof(value);
 
         /* Extract server path from registry. */
         sprintf(key, "CLSID\\%s\\LocalServer32", CLSID_VILEAUTO_KEY);
         if (RegOpenKeyEx(HKEY_CLASSES_ROOT,
-                         key,
+                         w32_charstring(key),
                          0,
                          KEY_QUERY_VALUE,
                          &hk) != ERROR_SUCCESS)
@@ -531,17 +631,17 @@ vile_oa::get_FullName(BSTR *pbstr)
             disp_win32_error(W32_SYS_ERROR, m_hwnd);
             return (E_UNEXPECTED);
         }
-        rc = RegQueryValueEx(hk, NULL, NULL, &regtype, value, &val_len);
+        rc = w32_get_reg_sz(hk, NULL, value, sizeof(value));
         RegCloseKey(hk);
         if (rc != ERROR_SUCCESS)
         {
             disp_win32_error(W32_SYS_ERROR, m_hwnd);
             return (E_UNEXPECTED);
         }
-        if ((cp = strchr((char *) value, ' ')) != NULL)
+        if ((cp = strchr(value, ' ')) != NULL)
             *cp = '\0';
-        tmp = TO_OLE_STRING((char *) value);
-        if (! (tmp && (m_bstrFullName = SysAllocString(tmp))))
+        tmp = TO_OLE_STRING(value);
+        if (! (tmp && (m_bstrFullName = SysAllocString(tmp)) != 0))
             return (E_OUTOFMEMORY);
     }
     *pbstr = SysAllocString(m_bstrFullName);
@@ -579,6 +679,7 @@ vile_oa::VileKeys(BSTR keys)
     HRESULT hr   = NOERROR;
     char    *msg = FROM_OLE_STRING(keys);
 
+    TRACE(("VileKeys %d bytes:\n%s\n", strlen(msg), msg));
     while (*msg)
     {
         if (! PostMessage(m_hwnd, WM_CHAR, *msg, 0))
@@ -639,6 +740,18 @@ vile_oa::get_InsertMode(VARIANT_BOOL *pbool)
     return NOERROR;
 }
 
+STDMETHODIMP
+vile_oa::get_GlobMode(VARIANT_BOOL *pbool)
+{
+    *pbool = (global_g_val(GMDGLOB)) ? VARIANT_TRUE : VARIANT_FALSE;
+    return NOERROR;
+}
+
+// ForegroundWindow() is not very useful on Win2K or XP...
+// See MSDN's description of SetForegroundWindow() for further details.
+// As an alternative, use vile_oa::get_MainHwnd() to obtain a handle
+// to the editor's main window and call SetForegoundWindow() directly.
+// See wvwrap.cpp for an example.
 STDMETHODIMP
 vile_oa::ForegroundWindow()
 {
@@ -648,6 +761,13 @@ vile_oa::ForegroundWindow()
     return NOERROR;
 }
 
+STDMETHODIMP
+vile_oa::get_MainHwnd(LONG *phwnd /* for VB compatibility */)
+{
+    *phwnd = (LONG) m_hwnd;   // ugh
+    return NOERROR;
+}
+
 STDMETHODIMP
 vile_oa::Minimize()
 {
@@ -742,6 +862,7 @@ waitfile(int f, int n)
 int
 syncfile(int f, int n)
 {
+#if VC6_ADDIN
     static CLSID     ds_clsid;   /* dev studio class id */
     HRESULT          hr;
     static int       initialized;
@@ -781,8 +902,8 @@ syncfile(int f, int n)
          */
 
         olebuf_len = ansibuf_len = 512;
-        ansibuf    = (char *) malloc(ansibuf_len);
-        olebuf     = (OLECHAR *) malloc(olebuf_len * sizeof(OLECHAR));
+        ansibuf    = typeallocn(char, ansibuf_len);
+        olebuf     = typeallocn(OLECHAR, olebuf_len);
         if (! (olebuf && ansibuf))
         {
             no_memory("syncfile()");
@@ -913,6 +1034,9 @@ syncfile(int f, int n)
 syncfile_exit:
     pApp->Release();
     return (rc);
+#else
+    return 0;			// FIXME
+#endif
 }
 
 /* ------------------- Key Redirection Implementation -------------------- */
@@ -1127,9 +1251,8 @@ invalid_keylist(void)
     ntwinio_redirect_hwnd(FALSE);
 
     /* Killed keyboard redirection, so make sure user sees error */
-    MessageBox((HWND) winvile_hwnd(),
+    w32_message_box((HWND) winvile_hwnd(),
                "invalid keylist syntax, redirection disabled",
-               prognam,
                MB_OK|MB_ICONSTOP);
     return (FALSE);
 }
@@ -1146,7 +1269,7 @@ unsupported_key(char *key)
     sprintf(msg, "VK_%s redir unsupported, redirection disabled", key);
 
     /* Killed keyboard redirection, so make sure user sees error */
-    MessageBox((HWND) winvile_hwnd(), msg, prognam, MB_OK|MB_ICONSTOP);
+    w32_message_box((HWND) winvile_hwnd(), msg, MB_OK|MB_ICONSTOP);
     return (FALSE);
 }
 
@@ -1159,7 +1282,7 @@ execute_action(int action)
 
     if (action & FLUSH_BUFFERS)
     {
-        rc = writeall(FALSE, 1, FALSE, FALSE, FALSE);
+        rc = writeall(FALSE, 1, FALSE, FALSE, FALSE, FALSE);
         update(FALSE);  /* move cursor out of mini-buffer */
 
         /*
@@ -1217,7 +1340,7 @@ execute_action(int action)
 static int
 build_keylist(void)
 {
-    char      *codestr, *redirlist, *tmp;
+    char      *redirlist, *tmp;
     REDIR_KEY *lcl_tbl;
 
     if (cache_tbl)
@@ -1364,7 +1487,8 @@ build_keylist(void)
 int
 oleauto_redirected_key(ULONG vk, ULONG modifier)
 {
-    int       attached, i, hit;
+    int       attached, hit;
+    unsigned  i;
     REDIR_KEY *lcl_tbl;
 
     if (vk >= TABLESIZE(keymap))
@@ -1444,7 +1568,7 @@ oleauto_redirected_key(ULONG vk, ULONG modifier)
                     redirect_tid);
 
             /* Killed keyboard redirection, so make sure user sees error */
-            MessageBox((HWND) winvile_hwnd(), msg, prognam, MB_OK|MB_ICONSTOP);
+            w32_message_box((HWND) winvile_hwnd(), msg, MB_OK|MB_ICONSTOP);
 
             /*
              * Return TRUE to signal the winvile message pump that key has
@@ -1463,7 +1587,7 @@ oleauto_redirected_key(ULONG vk, ULONG modifier)
                 redirect_hwnd);
 
         /* Killed keyboard redirection, so make sure user sees error */
-        MessageBox((HWND) winvile_hwnd(), msg, prognam, MB_OK|MB_ICONSTOP);
+        w32_message_box((HWND) winvile_hwnd(), msg, MB_OK|MB_ICONSTOP);
     }
     if (attached)
     {
