@@ -3,7 +3,7 @@
  * paragraph at a time.  There are all sorts of word mode commands.  If I
  * do any sentence mode commands, they are likely to be put in this file.
  *
- * $Header: /usr/build/vile/vile/RCS/word.c,v 1.71 1999/12/13 00:18:04 tom Exp $
+ * $Header: /usr/build/vile/vile/RCS/word.c,v 1.93 2009/05/17 23:33:39 tom Exp $
  *
  */
 
@@ -11,6 +11,11 @@
 #include	"edef.h"
 #include	"nefunc.h"
 
+#define each_character(c) \
+	    c = DOT.o, once = TRUE; \
+		(c > 0) || once; \
+		    c -= BytesBefore(DOT.l, c), once = FALSE
+
 /* Word wrap on n-spaces. Back-over whatever precedes the point on the current
  * line and stop on the first word-break or the beginning of the line. If we
  * reach the beginning of the line, jump back to the end of the word and start
@@ -21,79 +26,102 @@
 int
 wrapword(int f, int n)
 {
-    register int cnt = 0;	/* size of word wrapped to next line */
-    register int c;		/* character temporary */
-    B_COUNT to_delete;
+    int once;
+    int rc = FALSE;
+    int done = FALSE;
+    int cnt = 0;		/* size of word wrapped to next line */
+    int c;			/* character temporary */
+    B_COUNT to_delete = 0;
     C_NUM to_append = 0;
 
-    /* Back up from the <NL> 1 char */
-    if (!backchar(FALSE, 1))
-	return (FALSE);
+    TRACE((T_CALLED "wrapword(%d, %d)\n", f, n));
 
-    /* If parameter given, delete the trailing white space.  This is done
-     * to support a vi-like wrapmargin mode (see insert.c).  Unlike vi,
-     * we're deleting all of the trailing whitespace; vi deletes only the
-     * portion that was added during the current insertion.
-     *
-     * If there's no trailing whitespace, and we're not inserting blanks
-     * (n!=0), try to split the line at the last embedded whitespace.
-     */
-    if (f) {
-	register LINE *lp = DOT.l;
-	to_delete = 0L;
-	if (DOT.o >= 0 && !n && !isSpace(lgetc(lp, DOT.o))) {
-	    for (c = DOT.o; c >= 0; c--) {
+    /* Back up from the <NL> 1 char */
+    if (backchar(FALSE, 1)) {
+
+	/* If parameter given, delete the trailing white space.  This is done
+	 * to support a vi-like wrapmargin mode (see insert.c).  Unlike vi,
+	 * we're deleting all of the trailing whitespace; vi deletes only the
+	 * portion that was added during the current insertion.
+	 *
+	 * If there's no trailing whitespace, and we're not inserting blanks
+	 * (n!=0), try to split the line at the last embedded whitespace.
+	 */
+	if (f) {
+	    LINE *lp = DOT.l;
+	    to_delete = 0L;
+	    if (DOT.o >= 0 && !n && !isSpace(lgetc(lp, DOT.o))) {
+		int tmp = 0;
+		for (each_character(c)) {
+		    if (isSpace(lgetc(lp, c))) {
+			to_append = n;
+			cnt = tmp;
+			DOT.o = c;
+			break;
+		    }
+		    ++tmp;
+		}
+	    }
+	    for (each_character(c)) {
 		if (isSpace(lgetc(lp, c))) {
-		    to_append = n;
-		    cnt = (DOT.o - c);
+		    to_delete++;
 		    DOT.o = c;
+		} else {
 		    break;
 		}
 	    }
-	}
-	for (c = DOT.o; c >= 0; c--) {
-	    if (isSpace(lgetc(lp, c))) {
-		to_delete++;
-		DOT.o = c;
-	    } else {
-		break;
+	    if (to_delete == 0) {
+		DOT.o += BytesAt(DOT.l, DOT.o);
+	    }
+	} else {
+	    /* Back up until we aren't in a word, make sure there is a
+	     * break in the line
+	     */
+	    while (c = char_at(DOT), !isSpace(c)) {
+		cnt++;
+		if (!backchar(FALSE, 1)) {
+		    rc = FALSE;
+		    done = TRUE;
+		    break;
+		}
+	    }
+	    if (!done) {
+		/* if we make it to the beginning, start a new line */
+		if (DOT.o == 0) {
+		    (void) gotoeol(FALSE, 0);
+		    rc = lnewline();
+		    done = TRUE;
+		} else {
+		    to_delete = 1L;
+		}
 	    }
 	}
-	if (to_delete == 0)
-	    DOT.o++;
-    } else {
-	/* Back up until we aren't in a word, make sure there is a
-	 * break in the line
-	 */
-	while (c = char_at(DOT), !isSpace(c)) {
-	    cnt++;
-	    if (!backchar(FALSE, 1))
-		return (FALSE);
-	    /* if we make it to the beginning, start a new line */
-	    if (DOT.o == 0) {
-		(void) gotoeol(FALSE, 0);
-		return (lnewline());
+
+	if (!done) {
+	    /* delete the forward white space */
+	    if (ldel_bytes(to_delete, FALSE)) {
+
+		/* put in newline */
+		if (newline(FALSE, 1)) {
+
+		    rc = TRUE;
+		    /* and past the first word */
+		    while (cnt-- > 0) {
+			if (forwchar(FALSE, 1) == FALSE) {
+			    rc = FALSE;
+			    break;
+			}
+		    }
+		    if (rc) {
+			if (to_append > 0) {
+			    lins_bytes(to_append, ' ');
+			}
+		    }
+		}
 	    }
 	}
-	to_delete = 1L;
     }
-
-    /* delete the forward white space */
-    if (!ldelete(to_delete, FALSE))
-	return (FALSE);
-
-    /* put in newline */
-    if (!newline(FALSE, 1))
-	return FALSE;
-
-    /* and past the first word */
-    while (cnt-- > 0) {
-	if (forwchar(FALSE, 1) == FALSE)
-	    return (FALSE);
-    }
-    if (to_append > 0)
-	linsert(to_append, ' ');
-    return (TRUE);
+    returnCode(rc);
 }
 
 /*
@@ -120,7 +148,7 @@ forwviword(int f, int n)
     while (n--) {
 	int any = 0;
 	while (((s = isnewviwordf()) == FALSE) ||
-	    (s == SORTOFTRUE && n != 0)) {
+	       (s == SORTOFTRUE && n != 0)) {
 	    if (forwchar(TRUE, 1) == FALSE)
 		return (any != 0);
 	    any++;
@@ -148,7 +176,7 @@ forwword(int f, int n)
     while (n--) {
 	int any = 0;
 	while (((s = isnewwordf()) == FALSE) ||
-	    (s == SORTOFTRUE && n != 0)) {
+	       (s == SORTOFTRUE && n != 0)) {
 	    if (forwchar(TRUE, 1) == FALSE)
 		return (any != 0);
 	    any++;
@@ -167,9 +195,10 @@ int
 forwviendw(int f, int n)
 {
     int s = FALSE;
-    if (!f)
-	n = 1;
-    else if (n <= 0)
+
+    n = need_a_count(f, n, 1);
+
+    if (n <= 0)
 	return (FALSE);
     if (forwchar(TRUE, 1) == FALSE)
 	return (FALSE);
@@ -199,9 +228,10 @@ int
 forwendw(int f, int n)
 {
     int s = FALSE;
-    if (!f)
-	n = 1;
-    else if (n <= 0)
+
+    n = need_a_count(f, n, 1);
+
+    if (n <= 0)
 	return (FALSE);
     if (forwchar(TRUE, 1) == FALSE)
 	return (FALSE);
@@ -273,11 +303,14 @@ backword(int f, int n)
     return (forwchar(TRUE, 1));
 }
 
-int
-joinregion(void)
+/*
+ * If not 'exact', will adjust whitespace at the join-position.
+ */
+static int
+join_region(int exact)
 {
-    register int status;
-    register int doto, c;
+    int status;
+    int doto, c;
     LINE *end;
     REGION region;
     int done = FALSE;
@@ -287,7 +320,12 @@ joinregion(void)
 
 	DOT = region.r_orig;
 	end = region.r_end.l;
-	regionshape = EXACT;
+	if (DOT.l == end) {
+	    /* Degenerate case -- nothing to join */
+
+	    return (TRUE);
+	}
+	regionshape = rgn_EXACT;
 
 	while (!done && status == TRUE) {
 	    c = EOS;
@@ -297,14 +335,18 @@ joinregion(void)
 	    (void) setmark();
 	    if (forwline(FALSE, 1) != TRUE)
 		break;
-	    (void) firstnonwhite(FALSE, 1);
+	    if (exact) {
+		(void) gotobol(FALSE, 1);
+	    } else {
+		(void) firstnonwhite(FALSE, 1);
+	    }
 
 	    done = ((DOT.l == end) || (lforw(DOT.l) == end));
 	    if (killregionmaybesave(FALSE) != TRUE)
 		break;
 
 	    doto = DOT.o;
-	    if (doto == 0) {
+	    if (doto == 0 || exact) {
 		/*EMPTY */ ;
 	    }
 	    /* join at column 0 to empty line */
@@ -314,9 +356,9 @@ joinregion(void)
 		}
 		/* join after parentheses */
 		else if (lgetc(DOT.l, doto - 1) == '.') {
-		    status = linsert(2, ' ');
+		    status = lins_bytes(2, ' ');
 		} else if (!isSpace(c)) {
-		    status = linsert(1, ' ');
+		    status = lins_bytes(1, ' ');
 		}
 	    }
 	}
@@ -324,6 +366,18 @@ joinregion(void)
     return status;
 }
 
+int
+joinregion(void)
+{
+    return join_region(FALSE);
+}
+
+int
+joinregion_x(void)
+{
+    return join_region(TRUE);
+}
+
 int
 joinlines(int f, int n)
 {
@@ -339,7 +393,7 @@ dot_at_section_break(void)
     regexp *expC = b_val_rexp(curbp, VAL_COMMENTS)->reg;
 
     return (lregexec(expP, DOT.l, 0, llength(DOT.l)) ||
-	lregexec(expC, DOT.l, 0, llength(DOT.l)));
+	    lregexec(expC, DOT.l, 0, llength(DOT.l)));
 }
 
 /* returns the length of the comment-prefix, if it matches, otherwise -1 */
@@ -349,11 +403,55 @@ comment_prefix(void)
     regexp *expP = b_val_rexp(curbp, VAL_CMT_PREFIX)->reg;
     int result = -1;
     if (lregexec(expP, DOT.l, 0, llength(DOT.l))) {
-	result = (int) (expP->endp[0] - DOT.l->l_text);
+	result = (int) (expP->endp[0] - lvalue(DOT.l));
     }
     return result;
 }
 
+#if OPT_MULTIBYTE
+static int
+byte_to_columns(char text, int size)
+{
+    int result = 0;
+
+    if (b_is_utfXX(curbp)) {
+	result = (mb_cellwidth(curwp, &text, 1) * size);
+    } else {
+	result = size;
+    }
+    return result;
+}
+
+static int
+string_to_columns(TBUFF *text)
+{
+    long bytes = (long) tb_length(text);
+    int result = 0;
+
+    if (b_is_utfXX(curbp)) {
+	char *value = tb_values(text);
+
+	while (bytes > 0) {
+	    int step = vl_conv_to_utf32((UINT *) 0, value, (B_COUNT) bytes);
+	    if (step <= 0)
+		break;
+	    result += mb_cellwidth(curwp, value, step);
+	    value += step;
+	    bytes -= step;
+	}
+    } else {
+	result = bytes;
+    }
+
+    return result;
+}
+#define Byte2Column(s,n) byte_to_columns(s, n)
+#define String2Columns(s) string_to_columns(s)
+#else
+#define Byte2Column(s,n) n
+#define String2Columns(s) tb_length(s)
+#endif
+
 #define cplus_comment_start(c) \
 		((c == '/') \
 		&& DOT.o+1 < llength(DOT.l) \
@@ -364,11 +462,18 @@ comment_prefix(void)
 		&& DOT.o+1 < llength(DOT.l) \
 		&& lgetc(DOT.l,DOT.o+1) == '*')
 
-#define fmt_insert(count,chr) \
-		if ((s = linsert(count,chr)) != TRUE) \
+#define fmt_insert_chr(count,chr) \
+		if ((s = lins_bytes(count,chr)) != TRUE) \
 			return s; \
 		else \
-			clength += count
+			clength += Byte2Column(chr, count)
+
+#define fmt_insert_str(str) \
+		for (i = 0; i < tb_length(str); ++i) { \
+		    if ((s = lins_bytes(1, tb_values(str)[i])) != TRUE) \
+			return s; \
+		} \
+		clength += String2Columns(str)
 
 #define fmt_c_preprocessor(cp) \
 		(tb_length(*cp) == 1 \
@@ -380,19 +485,19 @@ comment_prefix(void)
 		&& memcmp("/*", tb_values(*wp), 2) == 0)
 
 static int
-do_formatting(TBUFF ** wp, TBUFF ** cp)
+do_formatting(TBUFF **wp, TBUFF **cp)
 {
-    register int c;		/* current char during scan     */
-    register int clength;	/* position on line during fill */
-    register int plength;	/* prefix to ignore during fill */
-    register ALLOC_T i;		/* index during word copy       */
-    register int newlen;	/* tentative new line length    */
-    register int finished;	/* Are we at the End-Of-Paragraph? */
-    register int firstflag;	/* first word? (needs no space) */
-    register int is_comment;	/* doing a comment block?       */
-    register int at_nl = TRUE;	/* just saw a newline?          */
-    register LINEPTR pastline;	/* pointer to line just past EOP */
-    register int sentence;	/* was the last char a period?  */
+    int c;			/* current char during scan     */
+    int clength;		/* position on line during fill */
+    int plength;		/* prefix to ignore during fill */
+    size_t i;			/* index during word copy       */
+    int newlen;			/* tentative new line length    */
+    int finished;		/* Are we at the End-Of-Paragraph? */
+    int firstflag;		/* first word? (needs no space) */
+    int is_comment;		/* doing a comment block?       */
+    int at_nl = TRUE;		/* just saw a newline?          */
+    LINE *pastline;		/* pointer to line just past EOP */
+    int sentence;		/* was the last char a period?  */
     int secondindent;
     int s;
     int fillcolumn = getfillcol(curbp);
@@ -443,8 +548,8 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 	if ((plength = comment_prefix()) >= 0) {
 	    is_comment = TRUE;
 	    tb_bappend(cp,
-		DOT.l->l_text + DOT.o,
-		(ALLOC_T) (plength - DOT.o));
+		       lvalue(DOT.l) + DOT.o,
+		       (size_t) (plength - DOT.o));
 	} else if (cplus_comment_start(c)) {
 	    is_comment = TRUE;
 	    tb_bappend(cp, "//", 2);
@@ -481,8 +586,8 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 		    if (s != (int) tb_length(*cp)
 			|| (s > 0
 			    && memcmp(tb_values(*cp),
-				DOT.l->l_text + DOT.o,
-				s))) {
+				      lvalue(DOT.l) + DOT.o,
+				      (size_t) s))) {
 			finished = SORTOFTRUE;
 		    }
 
@@ -491,14 +596,14 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 			tb_init(cp, EOS);
 			if (plength > 0) {
 			    tb_bappend(cp,
-				DOT.l->l_text + DOT.o,
-				(ALLOC_T) (plength));
+				       lvalue(DOT.l) + DOT.o,
+				       (size_t) (plength));
 			}
 			if (DOT.l != pastline
 			    && !dot_at_section_break()) {
-			    int spcs = DOT.o;
+			    B_COUNT spcs = (B_COUNT) DOT.o;
 			    DOT.o = 0;
-			    s = ldelete(spcs, FALSE);
+			    s = ldel_bytes(spcs, FALSE);
 			    if (s != TRUE)
 				return s;
 			}
@@ -517,7 +622,7 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 
 	    /* and then delete it */
 	    if (finished == FALSE) {
-		s = ldelete(1L, FALSE);
+		s = ldel_bytes(1L, FALSE);
 		if (s != TRUE)
 		    return s;
 	    }
@@ -531,7 +636,7 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 		    if (c == '?' || c == ':' || c == '!') {
 			sentence = TRUE;
 		    } else if (c == '.') {
-			unsigned len = tb_length(*wp);
+			size_t len = tb_length(*wp);
 			if (len == 4) {
 			    char *ptr = tb_values(*wp);
 			    /*
@@ -552,11 +657,11 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 		/* at a word break with a word waiting */
 		/* calculate tentative new length
 		   with word added */
-		newlen = clength + 1 + tb_length(*wp);
+		newlen = (int) (clength + 1 + String2Columns(*wp));
 		if (newlen <= fillcolumn) {
 		    /* add word to current line */
 		    if (!firstflag) {
-			fmt_insert(1, ' ');
+			fmt_insert_chr(1, ' ');
 		    }
 		} else {
 		    /* fix the leading indent now, if
@@ -566,28 +671,24 @@ do_formatting(TBUFF ** wp, TBUFF ** cp)
 		    if (lnewline() == FALSE)
 			return FALSE;
 		    clength = 0;
-		    fmt_insert(secondindent, ' ');
+		    fmt_insert_chr(secondindent, ' ');
 		    firstflag = TRUE;
 		}
 
 		if (firstflag
 		    && is_comment
 		    && !word_finishes_c_comment(wp)) {
-		    for (i = 0; i < tb_length(*cp); i++) {
-			fmt_insert(1, tb_values(*cp)[i]);
-		    }
+		    fmt_insert_str(*cp);
 		    if (!fmt_c_preprocessor(cp)) {
-			fmt_insert(1, ' ');
+			fmt_insert_chr(1, ' ');
 		    }
 		}
 		firstflag = FALSE;
 
 		/* and add the word in in either case */
-		for (i = 0; i < tb_length(*wp); i++) {
-		    fmt_insert(1, tb_values(*wp)[i]);
-		}
+		fmt_insert_str(*wp);
 		if (finished == FALSE && sentence) {
-		    fmt_insert(1, ' ');
+		    fmt_insert_chr(1, ' ');
 		}
 		tb_init(wp, EOS);
 	    }
@@ -627,12 +728,12 @@ formatregion(void)
 int
 wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 {
-    register LINE *lp;		/* current line to scan */
-    register int offset;	/* current char to scan */
-    long size;			/* size of region left to count */
-    register int ch;		/* current character to scan */
-    register int wordflag;	/* are we in a word now? */
-    register int lastflag;	/* were we just in a word? */
+    LINE *lp;			/* current line to scan */
+    int offset;			/* current char to scan */
+    ULONG size;			/* size of region left to count */
+    int ch;			/* current character to scan */
+    int wordflag;		/* are we in a word now? */
+    int lastflag;		/* were we just in a word? */
     long nwhite;		/* total # of blanks */
     long nwords;		/* total # of words */
     long nchars;		/* total number of chars */
@@ -641,7 +742,7 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
     int status;			/* status return code */
     REGION region;		/* region to look at */
     const char *ending = get_record_sep(curbp);
-    int len_rs = strlen(ending);
+    int len_rs = len_record_sep(curbp);
 
     /* make sure we have a region to count */
     if ((status = getregion(&region)) != TRUE) {
@@ -667,7 +768,7 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 	    ++nlines;
 	    if (len_rs > 1) {
 		nwhite += (len_rs - 1);
-		size -= (len_rs - 1);
+		size -= (ULONG) (len_rs - 1);
 	    }
 	} else {
 	    ch = lgetc(lp, offset);
@@ -691,8 +792,8 @@ wordcount(int f GCC_UNUSED, int n GCC_UNUSED)
 	avgch = 0.0;
 
     mlforce("lines %d, words %ld, chars %ld, avg chars/word %f",
-	nlines, nwords, nchars + nwhite, avgch);
+	    nlines, nwords, nchars + nwhite, avgch);
 
     return (TRUE);
 }
-#endif
+#endif /* OPT_WORDCOUNT */
